#ifndef _CONSTANTS_H_
#define _CONSTANTS_H_

// a place to put global constants so they do not have to be redefined

const long double PI = 3.1415926535897932384626433832795L;
const long double E = 2.718281828459045;

#endif /* _CONSTANTS_H_ */
